@@ -22,6 +22,7 @@
  */
 
 #include "si_pipe.h"
+#include "si_compute.h"
 #include "util/u_format.h"
 #include "util/u_surface.h"
 
@@ -120,6 +121,8 @@ si_blit_dbcb_copy(struct si_context *sctx,
 
 	assert(sctx->dbcb_depth_copy_enabled || sctx->dbcb_stencil_copy_enabled);
 
+	sctx->decompression_enabled = true;
+
 	while (level_mask) {
 		unsigned level = u_bit_scan(&level_mask);
 
@@ -163,6 +166,7 @@ si_blit_dbcb_copy(struct si_context *sctx,
 			fully_copied_levels |= 1u << level;
 	}
 
+	sctx->decompression_enabled = false;
 	sctx->dbcb_depth_copy_enabled = false;
 	sctx->dbcb_stencil_copy_enabled = false;
 	si_mark_atom_dirty(sctx, &sctx->db_render_state);
@@ -218,6 +222,8 @@ si_blit_decompress_zs_planes_in_place(struct si_context *sctx,
 
 	surf_tmpl.format = texture->resource.b.b.format;
 
+	sctx->decompression_enabled = true;
+
 	while (level_mask) {
 		unsigned level = u_bit_scan(&level_mask);
 
@@ -255,6 +261,7 @@ si_blit_decompress_zs_planes_in_place(struct si_context *sctx,
 	if (planes & PIPE_MASK_S)
 		texture->stencil_dirty_level_mask &= ~fully_decompressed_mask;
 
+	sctx->decompression_enabled = false;
 	sctx->db_flush_depth_inplace = false;
 	sctx->db_flush_stencil_inplace = false;
 	si_mark_atom_dirty(sctx, &sctx->db_render_state);
@@ -298,11 +305,11 @@ si_blit_decompress_zs_in_place(struct si_context *sctx,
 }
 
 static void
-si_flush_depth_texture(struct si_context *sctx,
-		       struct r600_texture *tex,
-		       unsigned required_planes,
-		       unsigned first_level, unsigned last_level,
-		       unsigned first_layer, unsigned last_layer)
+si_decompress_depth(struct si_context *sctx,
+		    struct r600_texture *tex,
+		    unsigned required_planes,
+		    unsigned first_level, unsigned last_level,
+		    unsigned first_layer, unsigned last_layer)
 {
 	unsigned inplace_planes = 0;
 	unsigned copy_planes = 0;
@@ -331,8 +338,6 @@ si_flush_depth_texture(struct si_context *sctx,
 		}
 	}
 
-	assert(!tex->tc_compatible_htile || levels_z == 0);
-
 	/* We may have to allocate the flushed texture here when called from
 	 * si_decompress_subresource.
 	 */
@@ -369,19 +374,47 @@ si_flush_depth_texture(struct si_context *sctx,
 	}
 
 	if (inplace_planes) {
-		si_blit_decompress_zs_in_place(
-			sctx, tex,
-			levels_z, levels_s,
-			first_layer, last_layer);
+		if (!tex->tc_compatible_htile) {
+			si_blit_decompress_zs_in_place(
+						sctx, tex,
+						levels_z, levels_s,
+						first_layer, last_layer);
+		}
+
+		/* Only in-place decompression needs to flush DB caches, or
+		 * when we don't decompress but TC-compatible planes are dirty.
+		 */
+		sctx->b.flags |= SI_CONTEXT_FLUSH_AND_INV_DB |
+				 SI_CONTEXT_INV_GLOBAL_L2 |
+				 SI_CONTEXT_INV_VMEM_L1;
+
+		/* If we flush DB caches for TC-compatible depth, the dirty
+		 * state becomes 0 for the whole mipmap tree and all planes.
+		 * (there is nothing else to flush)
+		 */
+		if (tex->tc_compatible_htile) {
+			if (r600_can_sample_zs(tex, false))
+				tex->dirty_level_mask = 0;
+			if (r600_can_sample_zs(tex, true))
+				tex->stencil_dirty_level_mask = 0;
+		}
+	}
+	/* set_framebuffer_state takes care of coherency for single-sample.
+	 * The DB->CB copy uses CB for the final writes.
+	 */
+	if (copy_planes && tex->resource.b.b.nr_samples > 1) {
+		sctx->b.flags |= SI_CONTEXT_INV_VMEM_L1 |
+				 SI_CONTEXT_INV_GLOBAL_L2 |
+				 SI_CONTEXT_FLUSH_AND_INV_CB;
 	}
 }
 
 static void
-si_flush_depth_textures(struct si_context *sctx,
-			struct si_textures_info *textures)
+si_decompress_sampler_depth_textures(struct si_context *sctx,
+				     struct si_textures_info *textures)
 {
 	unsigned i;
-	unsigned mask = textures->depth_texture_mask;
+	unsigned mask = textures->needs_depth_decompress_mask;
 
 	while (mask) {
 		struct pipe_sampler_view *view;
@@ -397,11 +430,10 @@ si_flush_depth_textures(struct si_context *sctx,
 		tex = (struct r600_texture *)view->texture;
 		assert(tex->db_compatible);
 
-		si_flush_depth_texture(
-				sctx, tex,
-				sview->is_stencil_sampler ? PIPE_MASK_S : PIPE_MASK_Z,
-				view->u.tex.first_level, view->u.tex.last_level,
-				0, util_max_layer(&tex->resource.b.b, view->u.tex.first_level));
+		si_decompress_depth(sctx, tex,
+				    sview->is_stencil_sampler ? PIPE_MASK_S : PIPE_MASK_Z,
+				    view->u.tex.first_level, view->u.tex.last_level,
+				    0, util_max_layer(&tex->resource.b.b, view->u.tex.first_level));
 	}
 }
 
@@ -427,16 +459,17 @@ static void si_blit_decompress_color(struct pipe_context *ctx,
 
 		/* disable levels without DCC */
 		for (int i = first_level; i <= last_level; i++) {
-			if (!rtex->dcc_offset ||
-			    i >= rtex->surface.num_dcc_levels)
+			if (!vi_dcc_enabled(rtex, i))
 				level_mask &= ~(1 << i);
 		}
 	} else if (rtex->fmask.size) {
-		custom_blend = sctx->custom_blend_decompress;
+		custom_blend = sctx->custom_blend_fmask_decompress;
 	} else {
-		custom_blend = sctx->custom_blend_fastclear;
+		custom_blend = sctx->custom_blend_eliminate_fastclear;
 	}
 
+	sctx->decompression_enabled = true;
+
 	while (level_mask) {
 		unsigned level = u_bit_scan(&level_mask);
 
@@ -454,10 +487,19 @@ static void si_blit_decompress_color(struct pipe_context *ctx,
 			surf_tmpl.u.tex.last_layer = layer;
 			cbsurf = ctx->create_surface(ctx, &rtex->resource.b.b, &surf_tmpl);
 
+			/* Required before and after FMASK and DCC_DECOMPRESS. */
+			if (custom_blend == sctx->custom_blend_fmask_decompress ||
+			    custom_blend == sctx->custom_blend_dcc_decompress)
+				sctx->b.flags |= SI_CONTEXT_FLUSH_AND_INV_CB;
+
 			si_blitter_begin(ctx, SI_DECOMPRESS);
 			util_blitter_custom_color(sctx->blitter, cbsurf, custom_blend);
 			si_blitter_end(ctx);
 
+			if (custom_blend == sctx->custom_blend_fmask_decompress ||
+			    custom_blend == sctx->custom_blend_dcc_decompress)
+				sctx->b.flags |= SI_CONTEXT_FLUSH_AND_INV_CB;
+
 			pipe_surface_reference(&cbsurf, NULL);
 		}
 
@@ -467,6 +509,25 @@ static void si_blit_decompress_color(struct pipe_context *ctx,
 			rtex->dirty_level_mask &= ~(1 << level);
 		}
 	}
+
+	sctx->decompression_enabled = false;
+
+	sctx->b.flags |= SI_CONTEXT_FLUSH_AND_INV_CB |
+			 SI_CONTEXT_INV_GLOBAL_L2 |
+			 SI_CONTEXT_INV_VMEM_L1;
+}
+
+static void
+si_decompress_color_texture(struct si_context *sctx, struct r600_texture *tex,
+			    unsigned first_level, unsigned last_level)
+{
+	/* CMASK or DCC can be discarded and we can still end up here. */
+	if (!tex->cmask.size && !tex->fmask.size && !tex->dcc_offset)
+		return;
+
+	si_blit_decompress_color(&sctx->b.b, tex, first_level, last_level, 0,
+				 util_max_layer(&tex->resource.b.b, first_level),
+				 false);
 }
 
 static void
@@ -474,7 +535,7 @@ si_decompress_sampler_color_textures(struct si_context *sctx,
 				     struct si_textures_info *textures)
 {
 	unsigned i;
-	unsigned mask = textures->compressed_colortex_mask;
+	unsigned mask = textures->needs_color_decompress_mask;
 
 	while (mask) {
 		struct pipe_sampler_view *view;
@@ -486,14 +547,9 @@ si_decompress_sampler_color_textures(struct si_context *sctx,
 		assert(view);
 
 		tex = (struct r600_texture *)view->texture;
-		/* CMASK or DCC can be discarded and we can still end up here. */
-		if (!tex->cmask.size && !tex->fmask.size && !tex->dcc_offset)
-			continue;
 
-		si_blit_decompress_color(&sctx->b.b, tex,
-					 view->u.tex.first_level, view->u.tex.last_level,
-					 0, util_max_layer(&tex->resource.b.b, view->u.tex.first_level),
-					 false);
+		si_decompress_color_texture(sctx, tex, view->u.tex.first_level,
+					    view->u.tex.last_level);
 	}
 }
 
@@ -502,7 +558,7 @@ si_decompress_image_color_textures(struct si_context *sctx,
 				   struct si_images_info *images)
 {
 	unsigned i;
-	unsigned mask = images->compressed_colortex_mask;
+	unsigned mask = images->needs_color_decompress_mask;
 
 	while (mask) {
 		const struct pipe_image_view *view;
@@ -514,14 +570,44 @@ si_decompress_image_color_textures(struct si_context *sctx,
 		assert(view->resource->target != PIPE_BUFFER);
 
 		tex = (struct r600_texture *)view->resource;
-		if (!tex->cmask.size && !tex->fmask.size && !tex->dcc_offset)
+
+		si_decompress_color_texture(sctx, tex, view->u.tex.level,
+					    view->u.tex.level);
+	}
+}
+
+static void si_check_render_feedback_texture(struct si_context *sctx,
+					     struct r600_texture *tex,
+					     unsigned first_level,
+					     unsigned last_level,
+					     unsigned first_layer,
+					     unsigned last_layer)
+{
+	bool render_feedback = false;
+
+	if (!tex->dcc_offset)
+		return;
+
+	for (unsigned j = 0; j < sctx->framebuffer.state.nr_cbufs; ++j) {
+		struct r600_surface * surf;
+
+		if (!sctx->framebuffer.state.cbufs[j])
 			continue;
 
-		si_blit_decompress_color(&sctx->b.b, tex,
-					 view->u.tex.level, view->u.tex.level,
-					 0, util_max_layer(&tex->resource.b.b, view->u.tex.level),
-					 false);
+		surf = (struct r600_surface*)sctx->framebuffer.state.cbufs[j];
+
+		if (tex == (struct r600_texture *)surf->base.texture &&
+		    surf->base.u.tex.level >= first_level &&
+		    surf->base.u.tex.level <= last_level &&
+		    surf->base.u.tex.first_layer <= last_layer &&
+		    surf->base.u.tex.last_layer >= first_layer) {
+			render_feedback = true;
+			break;
+		}
 	}
+
+	if (render_feedback)
+		r600_texture_disable_dcc(&sctx->b, tex);
 }
 
 static void si_check_render_feedback_textures(struct si_context *sctx,
@@ -532,7 +618,6 @@ static void si_check_render_feedback_textures(struct si_context *sctx,
 	while (mask) {
 		const struct pipe_sampler_view *view;
 		struct r600_texture *tex;
-		bool render_feedback = false;
 
 		unsigned i = u_bit_scan(&mask);
 
@@ -541,27 +626,12 @@ static void si_check_render_feedback_textures(struct si_context *sctx,
 			continue;
 
 		tex = (struct r600_texture *)view->texture;
-		if (!tex->dcc_offset)
-			continue;
 
-		for (unsigned j = 0; j < sctx->framebuffer.state.nr_cbufs; ++j) {
-			struct r600_surface * surf;
-
-			if (!sctx->framebuffer.state.cbufs[j])
-				continue;
-
-			surf = (struct r600_surface*)sctx->framebuffer.state.cbufs[j];
-
-			if (tex == (struct r600_texture*)surf->base.texture &&
-			    surf->base.u.tex.level >= view->u.tex.first_level &&
-			    surf->base.u.tex.level <= view->u.tex.last_level &&
-			    surf->base.u.tex.first_layer <= view->u.tex.last_layer &&
-			    surf->base.u.tex.last_layer >= view->u.tex.first_layer)
-				render_feedback = true;
-		}
-
-		if (render_feedback)
-			r600_texture_disable_dcc(&sctx->b, tex);
+		si_check_render_feedback_texture(sctx, tex,
+						 view->u.tex.first_level,
+						 view->u.tex.last_level,
+						 view->u.tex.first_layer,
+						 view->u.tex.last_layer);
 	}
 }
 
@@ -573,7 +643,6 @@ static void si_check_render_feedback_images(struct si_context *sctx,
 	while (mask) {
 		const struct pipe_image_view *view;
 		struct r600_texture *tex;
-		bool render_feedback = false;
 
 		unsigned i = u_bit_scan(&mask);
 
@@ -582,26 +651,54 @@ static void si_check_render_feedback_images(struct si_context *sctx,
 			continue;
 
 		tex = (struct r600_texture *)view->resource;
-		if (!tex->dcc_offset)
+
+		si_check_render_feedback_texture(sctx, tex,
+						 view->u.tex.level,
+						 view->u.tex.level,
+						 view->u.tex.first_layer,
+						 view->u.tex.last_layer);
+	}
+}
+
+static void si_check_render_feedback_resident_textures(struct si_context *sctx)
+{
+	util_dynarray_foreach(&sctx->resident_tex_handles,
+			      struct si_texture_handle *, tex_handle) {
+		struct pipe_sampler_view *view;
+		struct r600_texture *tex;
+
+		view = (*tex_handle)->view;
+		if (view->texture->target == PIPE_BUFFER)
 			continue;
 
-		for (unsigned j = 0; j < sctx->framebuffer.state.nr_cbufs; ++j) {
-			struct r600_surface * surf;
+		tex = (struct r600_texture *)view->texture;
 
-			if (!sctx->framebuffer.state.cbufs[j])
-				continue;
+		si_check_render_feedback_texture(sctx, tex,
+						 view->u.tex.first_level,
+						 view->u.tex.last_level,
+						 view->u.tex.first_layer,
+						 view->u.tex.last_layer);
+	}
+}
 
-			surf = (struct r600_surface*)sctx->framebuffer.state.cbufs[j];
+static void si_check_render_feedback_resident_images(struct si_context *sctx)
+{
+	util_dynarray_foreach(&sctx->resident_img_handles,
+			      struct si_image_handle *, img_handle) {
+		struct pipe_image_view *view;
+		struct r600_texture *tex;
 
-			if (tex == (struct r600_texture*)surf->base.texture &&
-			    surf->base.u.tex.level == view->u.tex.level &&
-			    surf->base.u.tex.first_layer <= view->u.tex.last_layer &&
-			    surf->base.u.tex.last_layer >= view->u.tex.first_layer)
-				render_feedback = true;
-		}
+		view = &(*img_handle)->view;
+		if (view->resource->target == PIPE_BUFFER)
+			continue;
 
-		if (render_feedback)
-			r600_texture_disable_dcc(&sctx->b, tex);
+		tex = (struct r600_texture *)view->resource;
+
+		si_check_render_feedback_texture(sctx, tex,
+						 view->u.tex.level,
+						 view->u.tex.level,
+						 view->u.tex.first_layer,
+						 view->u.tex.last_layer);
 	}
 }
 
@@ -615,9 +712,49 @@ static void si_check_render_feedback(struct si_context *sctx)
 		si_check_render_feedback_images(sctx, &sctx->images[i]);
 		si_check_render_feedback_textures(sctx, &sctx->samplers[i]);
 	}
+
+	si_check_render_feedback_resident_images(sctx);
+	si_check_render_feedback_resident_textures(sctx);
+
 	sctx->need_check_render_feedback = false;
 }
 
+static void si_decompress_resident_textures(struct si_context *sctx)
+{
+	util_dynarray_foreach(&sctx->resident_tex_needs_color_decompress,
+			      struct si_texture_handle *, tex_handle) {
+		struct pipe_sampler_view *view = (*tex_handle)->view;
+		struct r600_texture *tex = (struct r600_texture *)view->texture;
+
+		si_decompress_color_texture(sctx, tex, view->u.tex.first_level,
+					    view->u.tex.last_level);
+	}
+
+	util_dynarray_foreach(&sctx->resident_tex_needs_depth_decompress,
+			      struct si_texture_handle *, tex_handle) {
+		struct pipe_sampler_view *view = (*tex_handle)->view;
+		struct si_sampler_view *sview = (struct si_sampler_view *)view;
+		struct r600_texture *tex = (struct r600_texture *)view->texture;
+
+		si_decompress_depth(sctx, tex,
+			sview->is_stencil_sampler ? PIPE_MASK_S : PIPE_MASK_Z,
+			view->u.tex.first_level, view->u.tex.last_level,
+			0, util_max_layer(&tex->resource.b.b, view->u.tex.first_level));
+	}
+}
+
+static void si_decompress_resident_images(struct si_context *sctx)
+{
+	util_dynarray_foreach(&sctx->resident_img_needs_color_decompress,
+			      struct si_image_handle *, img_handle) {
+		struct pipe_image_view *view = &(*img_handle)->view;
+		struct r600_texture *tex = (struct r600_texture *)view->resource;
+
+		si_decompress_color_texture(sctx, tex, view->u.tex.level,
+					    view->u.tex.level);
+	}
+}
+
 static void si_decompress_textures(struct si_context *sctx, unsigned shader_mask)
 {
 	unsigned compressed_colortex_counter, mask;
@@ -629,25 +766,37 @@ static void si_decompress_textures(struct si_context *sctx, unsigned shader_mask
 	compressed_colortex_counter = p_atomic_read(&sctx->screen->b.compressed_colortex_counter);
 	if (compressed_colortex_counter != sctx->b.last_compressed_colortex_counter) {
 		sctx->b.last_compressed_colortex_counter = compressed_colortex_counter;
-		si_update_compressed_colortex_masks(sctx);
+		si_update_needs_color_decompress_masks(sctx);
 	}
 
 	/* Decompress color & depth textures if needed. */
-	mask = sctx->compressed_tex_shader_mask & shader_mask;
+	mask = sctx->shader_needs_decompress_mask & shader_mask;
 	while (mask) {
 		unsigned i = u_bit_scan(&mask);
 
-		if (sctx->samplers[i].depth_texture_mask) {
-			si_flush_depth_textures(sctx, &sctx->samplers[i]);
+		if (sctx->samplers[i].needs_depth_decompress_mask) {
+			si_decompress_sampler_depth_textures(sctx, &sctx->samplers[i]);
 		}
-		if (sctx->samplers[i].compressed_colortex_mask) {
+		if (sctx->samplers[i].needs_color_decompress_mask) {
 			si_decompress_sampler_color_textures(sctx, &sctx->samplers[i]);
 		}
-		if (sctx->images[i].compressed_colortex_mask) {
+		if (sctx->images[i].needs_color_decompress_mask) {
 			si_decompress_image_color_textures(sctx, &sctx->images[i]);
 		}
 	}
 
+	if (shader_mask & u_bit_consecutive(0, SI_NUM_GRAPHICS_SHADERS)) {
+		if (sctx->uses_bindless_samplers)
+			si_decompress_resident_textures(sctx);
+		if (sctx->uses_bindless_images)
+			si_decompress_resident_images(sctx);
+	} else if (shader_mask & (1 << PIPE_SHADER_COMPUTE)) {
+		if (sctx->cs_shader_state.program->uses_bindless_samplers)
+			si_decompress_resident_textures(sctx);
+		if (sctx->cs_shader_state.program->uses_bindless_images)
+			si_decompress_resident_images(sctx);
+	}
+
 	si_check_render_feedback(sctx);
 }
 
@@ -700,7 +849,7 @@ static void si_clear(struct pipe_context *ctx, unsigned buffers,
 		}
 	}
 
-	if (zstex && zstex->htile_buffer &&
+	if (zstex && zstex->htile_offset &&
 	    zsbuf->u.tex.level == 0 &&
 	    zsbuf->u.tex.first_layer == 0 &&
 	    zsbuf->u.tex.last_layer == util_max_layer(&zstex->resource.b.b, 0)) {
@@ -813,10 +962,32 @@ static void si_decompress_subresource(struct pipe_context *ctx,
 		if (!(rtex->surface.flags & RADEON_SURF_SBUFFER))
 			planes &= ~PIPE_MASK_S;
 
-		si_flush_depth_texture(sctx, rtex, planes,
-				       level, level,
-				       first_layer, last_layer);
+		/* If we've rendered into the framebuffer and it's a blitting
+		 * source, make sure the decompression pass is invoked
+		 * by dirtying the framebuffer.
+		 */
+		if (sctx->framebuffer.state.zsbuf &&
+		    sctx->framebuffer.state.zsbuf->u.tex.level == level &&
+		    sctx->framebuffer.state.zsbuf->texture == tex)
+			si_update_fb_dirtiness_after_rendering(sctx);
+
+		si_decompress_depth(sctx, rtex, planes,
+				    level, level,
+				    first_layer, last_layer);
 	} else if (rtex->fmask.size || rtex->cmask.size || rtex->dcc_offset) {
+		/* If we've rendered into the framebuffer and it's a blitting
+		 * source, make sure the decompression pass is invoked
+		 * by dirtying the framebuffer.
+		 */
+		for (unsigned i = 0; i < sctx->framebuffer.state.nr_cbufs; i++) {
+			if (sctx->framebuffer.state.cbufs[i] &&
+			    sctx->framebuffer.state.cbufs[i]->u.tex.level == level &&
+			    sctx->framebuffer.state.cbufs[i]->texture == tex) {
+				si_update_fb_dirtiness_after_rendering(sctx);
+				break;
+			}
+		}
+
 		si_blit_decompress_color(ctx, rtex, level, level,
 					 first_layer, last_layer, false);
 	}
@@ -845,7 +1016,7 @@ void si_resource_copy_region(struct pipe_context *ctx,
 	struct pipe_surface *dst_view, dst_templ;
 	struct pipe_sampler_view src_templ, *src_view;
 	unsigned dst_width, dst_height, src_width0, src_height0;
-	unsigned src_force_level = 0;
+	unsigned dst_width0, dst_height0, src_force_level = 0;
 	struct pipe_box sbox, dstbox;
 
 	/* Handle buffers first. */
@@ -863,11 +1034,13 @@ void si_resource_copy_region(struct pipe_context *ctx,
 
 	dst_width = u_minify(dst->width0, dst_level);
 	dst_height = u_minify(dst->height0, dst_level);
+	dst_width0 = dst->width0;
+	dst_height0 = dst->height0;
 	src_width0 = src->width0;
 	src_height0 = src->height0;
 
 	util_blitter_default_dst_texture(&dst_templ, dst, dst_level, dstz);
-	util_blitter_default_src_texture(&src_templ, src, src_level);
+	util_blitter_default_src_texture(sctx->blitter, &src_templ, src, src_level);
 
 	if (util_format_is_compressed(src->format) ||
 	    util_format_is_compressed(dst->format)) {
@@ -881,6 +1054,8 @@ void si_resource_copy_region(struct pipe_context *ctx,
 
 		dst_width = util_format_get_nblocksx(dst->format, dst_width);
 		dst_height = util_format_get_nblocksy(dst->format, dst_height);
+		dst_width0 = util_format_get_nblocksx(dst->format, dst_width0);
+		dst_height0 = util_format_get_nblocksy(dst->format, dst_height0);
 		src_width0 = util_format_get_nblocksx(src->format, src_width0);
 		src_height0 = util_format_get_nblocksy(src->format, src_height0);
 
@@ -896,14 +1071,13 @@ void si_resource_copy_region(struct pipe_context *ctx,
 		src_box = &sbox;
 
 		src_force_level = src_level;
-	} else if (!util_blitter_is_copy_supported(sctx->blitter, dst, src) ||
-		   /* also *8_SNORM has precision issues, use UNORM instead */
-		   util_format_is_snorm8(src->format)) {
+	} else if (!util_blitter_is_copy_supported(sctx->blitter, dst, src)) {
 		if (util_format_is_subsampled_422(src->format)) {
 			src_templ.format = PIPE_FORMAT_R8G8B8A8_UINT;
 			dst_templ.format = PIPE_FORMAT_R8G8B8A8_UINT;
 
 			dst_width = util_format_get_nblocksx(dst->format, dst_width);
+			dst_width0 = util_format_get_nblocksx(dst->format, dst_width0);
 			src_width0 = util_format_get_nblocksx(src->format, src_width0);
 
 			dstx = util_format_get_nblocksx(dst->format, dstx);
@@ -944,8 +1118,51 @@ void si_resource_copy_region(struct pipe_context *ctx,
 		}
 	}
 
+	/* SNORM8 blitting has precision issues on some chips. Use the SINT
+	 * equivalent instead, which doesn't force DCC decompression.
+	 * Note that some chips avoid this issue by using SDMA.
+	 */
+	if (util_format_is_snorm8(dst_templ.format)) {
+		switch (dst_templ.format) {
+		case PIPE_FORMAT_R8_SNORM:
+			dst_templ.format = src_templ.format = PIPE_FORMAT_R8_SINT;
+			break;
+		case PIPE_FORMAT_R8G8_SNORM:
+			dst_templ.format = src_templ.format = PIPE_FORMAT_R8G8_SINT;
+			break;
+		case PIPE_FORMAT_R8G8B8X8_SNORM:
+			dst_templ.format = src_templ.format = PIPE_FORMAT_R8G8B8X8_SINT;
+			break;
+		case PIPE_FORMAT_R8G8B8A8_SNORM:
+		/* There are no SINT variants for ABGR and XBGR, so we have to use RGBA. */
+		case PIPE_FORMAT_A8B8G8R8_SNORM:
+		case PIPE_FORMAT_X8B8G8R8_SNORM:
+			dst_templ.format = src_templ.format = PIPE_FORMAT_R8G8B8A8_SINT;
+			break;
+		case PIPE_FORMAT_A8_SNORM:
+			dst_templ.format = src_templ.format = PIPE_FORMAT_A8_SINT;
+			break;
+		case PIPE_FORMAT_L8_SNORM:
+			dst_templ.format = src_templ.format = PIPE_FORMAT_L8_SINT;
+			break;
+		case PIPE_FORMAT_L8A8_SNORM:
+			dst_templ.format = src_templ.format = PIPE_FORMAT_L8A8_SINT;
+			break;
+		case PIPE_FORMAT_I8_SNORM:
+			dst_templ.format = src_templ.format = PIPE_FORMAT_I8_SINT;
+			break;
+		default:; /* fall through */
+		}
+	}
+
+	vi_disable_dcc_if_incompatible_format(&sctx->b, dst, dst_level,
+					      dst_templ.format);
+	vi_disable_dcc_if_incompatible_format(&sctx->b, src, src_level,
+					      src_templ.format);
+
 	/* Initialize the surface. */
 	dst_view = r600_create_surface_custom(ctx, dst, &dst_templ,
+					      dst_width0, dst_height0,
 					      dst_width, dst_height);
 
 	/* Initialize the sampler view. */
@@ -968,6 +1185,29 @@ void si_resource_copy_region(struct pipe_context *ctx,
 	pipe_sampler_view_reference(&src_view, NULL);
 }
 
+static void si_do_CB_resolve(struct si_context *sctx,
+			     const struct pipe_blit_info *info,
+			     struct pipe_resource *dst,
+			     unsigned dst_level, unsigned dst_z,
+			     enum pipe_format format)
+{
+	/* Required before and after CB_RESOLVE. */
+	sctx->b.flags |= SI_CONTEXT_FLUSH_AND_INV_CB;
+
+	si_blitter_begin(&sctx->b.b, SI_COLOR_RESOLVE |
+			 (info->render_condition_enable ? 0 : SI_DISABLE_RENDER_COND));
+	util_blitter_custom_resolve_color(sctx->blitter, dst, dst_level, dst_z,
+					  info->src.resource, info->src.box.z,
+					  ~0, sctx->custom_blend_resolve,
+					  format);
+	si_blitter_end(&sctx->b.b);
+
+	/* Flush caches for possible texturing. */
+	sctx->b.flags |= SI_CONTEXT_FLUSH_AND_INV_CB |
+			 SI_CONTEXT_INV_GLOBAL_L2 |
+			 SI_CONTEXT_INV_VMEM_L1;
+}
+
 static bool do_hardware_msaa_resolve(struct pipe_context *ctx,
 				     const struct pipe_blit_info *info)
 {
@@ -978,7 +1218,6 @@ static bool do_hardware_msaa_resolve(struct pipe_context *ctx,
 	unsigned dst_width = u_minify(info->dst.resource->width0, info->dst.level);
 	unsigned dst_height = u_minify(info->dst.resource->height0, info->dst.level);
 	enum pipe_format format = info->src.format;
-	unsigned sample_mask = ~0;
 	struct pipe_resource *tmp, templ;
 	struct pipe_blit_info blit;
 
@@ -1033,23 +1272,20 @@ static bool do_hardware_msaa_resolve(struct pipe_context *ctx,
 		 * it's being overwritten anyway, clear it to uncompressed.
 		 * This is still the fastest codepath even with this clear.
 		 */
-		if (dst->dcc_offset &&
-		    info->dst.level < dst->surface.num_dcc_levels) {
+		if (vi_dcc_enabled(dst, info->dst.level)) {
+			/* TODO: Implement per-level DCC clears for GFX9. */
+			if (sctx->b.chip_class >= GFX9 &&
+			    info->dst.resource->last_level != 0)
+				goto resolve_to_temp;
+
 			vi_dcc_clear_level(&sctx->b, dst, info->dst.level,
 					   0xFFFFFFFF);
 			dst->dirty_level_mask &= ~(1 << info->dst.level);
 		}
 
 		/* Resolve directly from src to dst. */
-		si_blitter_begin(ctx, SI_COLOR_RESOLVE |
-				 (info->render_condition_enable ? 0 : SI_DISABLE_RENDER_COND));
-		util_blitter_custom_resolve_color(sctx->blitter,
-						  info->dst.resource, info->dst.level,
-						  info->dst.box.z,
-						  info->src.resource, info->src.box.z,
-						  sample_mask, sctx->custom_blend_resolve,
-						  format);
-		si_blitter_end(ctx);
+		si_do_CB_resolve(sctx, info, info->dst.resource,
+				 info->dst.level, info->dst.box.z, format);
 		return true;
 	}
 
@@ -1083,13 +1319,7 @@ resolve_to_temp:
 	assert(src->surface.micro_tile_mode == rtmp->surface.micro_tile_mode);
 
 	/* resolve */
-	si_blitter_begin(ctx, SI_COLOR_RESOLVE |
-			 (info->render_condition_enable ? 0 : SI_DISABLE_RENDER_COND));
-	util_blitter_custom_resolve_color(sctx->blitter, tmp, 0, 0,
-					  info->src.resource, info->src.box.z,
-					  sample_mask, sctx->custom_blend_resolve,
-					  format);
-	si_blitter_end(ctx);
+	si_do_CB_resolve(sctx, info, tmp, 0, 0, format);
 
 	/* blit */
 	blit = *info;
@@ -1136,10 +1366,10 @@ static void si_blit(struct pipe_context *ctx,
 
 	/* The driver doesn't decompress resources automatically while
 	 * u_blitter is rendering. */
-	vi_dcc_disable_if_incompatible_format(&sctx->b, info->src.resource,
+	vi_disable_dcc_if_incompatible_format(&sctx->b, info->src.resource,
 					      info->src.level,
 					      info->src.format);
-	vi_dcc_disable_if_incompatible_format(&sctx->b, info->dst.resource,
+	vi_disable_dcc_if_incompatible_format(&sctx->b, info->dst.resource,
 					      info->dst.level,
 					      info->dst.format);
 	si_decompress_subresource(ctx, info->src.resource, info->mask,
@@ -1171,7 +1401,7 @@ static boolean si_generate_mipmap(struct pipe_context *ctx,
 
 	/* The driver doesn't decompress resources automatically while
 	 * u_blitter is rendering. */
-	vi_dcc_disable_if_incompatible_format(&sctx->b, tex, base_level,
+	vi_disable_dcc_if_incompatible_format(&sctx->b, tex, base_level,
 					      format);
 	si_decompress_subresource(ctx, tex, PIPE_MASK_RGBAZS,
 				  base_level, first_layer, last_layer);
@@ -1181,11 +1411,15 @@ static boolean si_generate_mipmap(struct pipe_context *ctx,
 	rtex->dirty_level_mask &= ~u_bit_consecutive(base_level + 1,
 						     last_level - base_level);
 
+	sctx->generate_mipmap_for_depth = rtex->is_depth;
+
 	si_blitter_begin(ctx, SI_BLIT | SI_DISABLE_RENDER_COND);
 	util_blitter_generate_mipmap(sctx->blitter, tex, format,
 				     base_level, last_level,
 				     first_layer, last_layer);
 	si_blitter_end(ctx);
+
+	sctx->generate_mipmap_for_depth = false;
 	return true;
 }
 
